use camino::{Utf8Path, Utf8PathBuf};
use color_eyre::Result;
use color_eyre::eyre::{OptionExt, eyre};
use http_body_util::BodyExt;
use hyper::{Request, StatusCode};
use hyper_util::rt::TokioIo;
use serde_json::json;
use std::process::Stdio;
use tokio::io::AsyncWriteExt;
use tokio::net::UnixStream;
use tokio::process::Command;

#[derive(Clone)]
pub struct CaddyController {
    caddy_admin_socket_path: Utf8PathBuf,
    api_socket_path: Utf8PathBuf,
}

impl CaddyController {
    pub async fn upsert_site_configuration(
        &self,
        domain: &String,
        content_path: &Utf8Path,
    ) -> Result<()> {
        let configuration_object = json!({
            "@id": domain,
            "match": [{ "host": [domain] }],
            "handle": [
                {
                    "group": "1",
                    "match": [{ "path": ["/sscdc/*"] }],
                    "handle": [
                        {
                            "handler": "rewrite",
                            "strip_path_prefix": "/_sscdc"
                        },
                        {
                            "handler": "reverse_proxy",
                            "upstreams": [{ "dial": format!("unix/{}", &self.api_socket_path) }],
                        }
                    ]
                },
                {
                    "group": "1",
                    "handle": [
                        { "handler": "vars", "root": content_path.to_string() },
                        {
                            "handler": "file_server",
                            "precompressed": {
                                "br": {},
                                "gzip": {}
                            },
                            "precompressed_order": ["br", "gzip"]
                        }
                    ]
                }
            ],
            "terminal": true
        });

        let stream = UnixStream::connect(&self.caddy_admin_socket_path).await?;
        let (mut sender, connection) =
            hyper::client::conn::http1::handshake(TokioIo::new(stream)).await?;

        tokio::task::spawn(async move { connection.await.unwrap() });

        let response = sender
            .send_request(
                Request::builder()
                    .uri(&format!("/id/{}", domain))
                    .body(configuration_object.to_string().boxed())?,
            )
            .await?;

        match response.status() {
            StatusCode::OK => {}
            StatusCode::NOT_FOUND => {
                // The site does not yet exist.
                let response = sender
                    .send_request(
                        Request::builder()
                            .uri("/config/apps/http/servers/srv0/routes")
                            .body(configuration_object.to_string().boxed())?,
                    )
                    .await?;

                let status = response.status();

                if !status.is_success() {
                    return Err(eyre!(
                        "The configuration update request to Caddy failed with status code {}",
                        status
                    ));
                }
            }
            c => {
                return Err(eyre!(
                    "The configuration update request to Caddy failed with status code {}",
                    c
                ));
            }
        }

        Ok(())
    }
}

fn get_initial_caddy_configuration_object(
    admin_api_socket_path: &Utf8Path,
    api_socket_path: &Utf8Path,
) -> serde_json::Value {
    json!({
        "admin": {
            "listen": format!("unix/{}", admin_api_socket_path),
            "config": { "persist": false }
        },
        "logging": {
            "sink": {
                "writer": { "output": "discard" }
            },
            "logs": {
                "": {
                    "writer": { "output": "discard" }
                }
            }
        },
        "storage": {
            "module": "file_system",
            "root": "/tmp/sscdc-caddy"
        },
        "apps": {
            "http": {
                "servers": {
                    "srv0": {
                        "listen": [":80"],
                        "routes": [{
                            "match": [{ "path": ["/sscdc/*"] }],
                            "handle": [
                                {
                                    "handler": "rewrite",
                                    "strip_path_prefix": "/_sscdc"
                                },
                                {
                                    "handler": "reverse_proxy",
                                    "upstreams": [{ "dial": format!("unix/{}", api_socket_path) }],
                                }
                            ]
                        }],
                        "automatic_https": { "disable": true }
                    }
                }
            }
        }
    })
}

pub async fn start_caddy(
    api_socket_path: &Utf8Path,
    sockets_directory_path: &Utf8Path,
) -> Result<CaddyController> {
    let caddy_admin_socket_path = sockets_directory_path.join("caddy-admin-api.sock");
    let process = Command::new("caddy")
        .args(&["run", "--config", "-"])
        .stdin(Stdio::piped())
        .spawn()?;

    let initial_configuration_object =
        get_initial_caddy_configuration_object(&caddy_admin_socket_path, api_socket_path);

    let initial_configuration_string = initial_configuration_object.to_string();
    log::debug!(
        "Initial caddy configuration: {}",
        initial_configuration_string
    );

    process
        .stdin
        .ok_or_eyre("The caddy process stdin is not open.")?
        .write_all(initial_configuration_string.as_ref())
        .await?;

    Ok(CaddyController {
        caddy_admin_socket_path,
        api_socket_path: api_socket_path.to_path_buf(),
    })
}
