#![feature(never_type)]

use crate::api::start_api_server;
use crate::caddy::start_caddy;
use crate::config::load_config;
use crate::sites::start_sites_worker;
use camino::Utf8Path;
use color_eyre::Result;
use color_eyre::eyre::WrapErr;
use log::LevelFilter;
use std::fs;

mod api;
mod caddy;
mod config;
mod sites;

#[tokio::main(flavor = "current_thread")]
async fn main() -> Result<()> {
    color_eyre::install()?;
    env_logger::Builder::new()
        .filter_module("sscdc", LevelFilter::Info)
        .parse_default_env()
        .init();

    log::info!("Loading configuration…");
    let config = load_config()?;

    let sockets_directory_path = Utf8Path::new(&config.sockets_directory);
    fs::create_dir_all(&sockets_directory_path)
        .wrap_err("Failed to access or create the sockets directory.")?;
    let sockets_directory_path = sockets_directory_path.canonicalize_utf8().unwrap();

    log::info!("Starting internal API server…");
    let api_socket_path = sockets_directory_path.join("api.sock");
    start_api_server(&api_socket_path).await?;
    log::debug!("The internal API server is listening at {api_socket_path}");

    log::info!("Starting the reverse proxy…");
    let caddy_controller = start_caddy(&api_socket_path, &sockets_directory_path).await?;

    let sites_worker = start_sites_worker(config.sites_directory.to_string().into(), caddy_controller).await?;

    log::info!("Startup complete.");

    sites_worker.join_handle.await??;
}
