use crate::caddy::CaddyController;
use camino::{Utf8Path, Utf8PathBuf};
use color_eyre::Result;
use color_eyre::eyre::WrapErr;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::sync::{Arc, RwLockReadGuard, RwLockWriteGuard};
use tokio::fs;
use tokio::io::{AsyncWriteExt, BufWriter};
use tokio::sync::mpsc::{UnboundedReceiver, UnboundedSender};
use tokio::sync::{Mutex, RwLock};
use tokio::task::JoinHandle;
use futures::StreamExt;

#[derive(Serialize, Deserialize)]
pub struct SiteState {
    active_version: Option<SiteStateVersion>,
    current_version: Option<SiteStateVersion>,
}

impl SiteState {
    async fn read_from_site_directory(site_directory_path: &Utf8Path) -> Result<Self> {
        let string = fs::read_to_string(site_directory_path.join("site.toml")).await?;
        Ok(toml::from_str::<SiteState>(&string)?)
    }

    async fn write_to_site_directory(&self, site_directory_path: &Utf8Path) -> Result<()> {
        fs::write(
            site_directory_path.join("site.toml"),
            toml::to_string(&self).unwrap(),
        )
        .await?;
        Ok(())
    }

    fn is_outdated(&self) -> bool {
        self.get_current_version_if_outdated().is_some()
    }

    fn get_current_version_if_outdated(&self) -> Option<&SiteStateVersion> {
        if let Some(current_version) = &self.current_version {
            if let Some(active_version) = &self.active_version {
                if current_version.id != active_version.id {
                    return Some(&active_version)
                }
            }
        }

        None
    }
}

#[derive(Serialize, Deserialize, Clone)]
pub struct SiteStateVersion {
    id: String,
    download_url: String,
}

pub struct Site {
    pub domain: String,
    pub path: Utf8PathBuf,
    pub state: RwLock<SiteState>,
}

type Sites = RwLock<HashMap<String, Arc<Site>>>;

pub struct SitesWorker {
    sites: Arc<Sites>,
    sites_directory_path: Utf8PathBuf,
    download_tasks_sender: UnboundedSender<String>,
    pub join_handle: JoinHandle<Result<!>>,
}

impl SitesWorker {
    pub async fn set_current_version(
        &self,
        domain: String,
        id: String,
        download_url: String,
    ) -> Result<()> {
        let mut sites = self.sites.write().await;
        let site = sites.entry(domain.clone()).or_insert_with(|| Arc::new(Site {
            domain: domain.clone(),
            path: self.sites_directory_path.join(&domain),
            state: RwLock::new(SiteState {
                current_version: None,
                active_version: None,
            }),
        }));

        let mut state = site.state.write().await;
        state.current_version = Some(SiteStateVersion { id, download_url });
        state.write_to_site_directory(&site.path).await?;
        self.download_tasks_sender.send(domain).unwrap();

        Ok(())
    }
}

pub async fn load_sites(sites_directory_path: &Utf8Path) -> Result<Sites> {
    let mut sites: HashMap<String, Arc<Site>> = HashMap::new();

    if fs::try_exists(&sites_directory_path).await.unwrap_or(false) {
        let mut dir = fs::read_dir(&sites_directory_path).await?;
        loop {
            if let Some(entry) = dir.next_entry().await? {
                if entry.file_type().await?.is_dir() {
                    let path = Utf8PathBuf::from_path_buf(entry.path()).unwrap();
                    let domain = path.file_name().unwrap().to_string();
                    let state = SiteState::read_from_site_directory(&path).await?;

                    sites.insert(
                        domain.clone(),
                        Arc::new(Site {
                            path,
                            domain,
                            state: RwLock::new(state),
                        }),
                    );
                }
            } else {
                break;
            }
        }
    } else {
        fs::create_dir_all(&sites_directory_path)
            .await
            .wrap_err("Failed to create the sites directory (and its parents)")?;
    }

    Ok(RwLock::new(sites))
}

async fn handle_download(sites: &Sites, caddy_controller: &CaddyController, domain: String) -> Result<()> {
    let site = {
        let sites = sites.read().await;
        match sites.get(&domain) {
            None => return Ok(()),
            Some(a) => Arc::clone(a)
        }
    };

    let mut state = site.state.write().await;
    let current_version = match state.get_current_version_if_outdated() {
        None => return Ok(()),
        Some(v) => v
    };

    // Download
    let mut archive_file = fs::File::create(site.path.join(format!("{}.zip", current_version.id))).await?;
    let mut file_writer = BufWriter::new(&mut archive_file);

    let mut response = reqwest::get(&current_version.download_url).await?;
    
    while let Some(chunk) = response.chunk().await? {
        file_writer.write_all(&chunk).await?;
    }

    let extraction_directory_path = site.path.join(&current_version.id);
    let _ = fs::remove_dir_all(&extraction_directory_path).await;
    fs::create_dir_all(&extraction_directory_path).await?;
    
    // Unpack to temp dir
    tokio::task::spawn_blocking({
        let archive_file = archive_file.into_std().await;
        let extraction_directory_path = extraction_directory_path.clone();
        
        move || -> Result<()> {
            let mut archive = zip::ZipArchive::new(archive_file)?;
            archive.extract(&extraction_directory_path)?;
            
            Ok(())
        }
    }).await??;
    
    // Update and write state
    state.active_version = Some(current_version.clone());
    state.write_to_site_directory(&site.path).await?;
    
    // Update Caddy configuration
    caddy_controller.upsert_site_configuration(&domain, &extraction_directory_path).await?;

    Ok(())
}

async fn sites_worker(
    mut download_tasks_receiver: UnboundedReceiver<String>,
    sites: Arc<Sites>,
    caddy_controller: CaddyController,
) -> Result<!> {
    loop {
        let domain = download_tasks_receiver.recv().await.unwrap();
        handle_download(&sites, &caddy_controller, domain).await?;
    }
}

pub async fn start_sites_worker(
    sites_directory_path: Utf8PathBuf,
    caddy_controller: CaddyController,
) -> Result<SitesWorker> {
    let (download_tasks_sender, download_tasks_receiver) = tokio::sync::mpsc::unbounded_channel();
    let sites = Arc::new(load_sites(&sites_directory_path).await?);

    for (_, site) in sites.read().await.iter() {
        let state = site.state.read().await;
        if let Some(current_version) = &state.current_version {
            if let Some(active_version) = &state.active_version {
                if current_version.id != active_version.id {
                    download_tasks_sender.send(site.domain.clone()).unwrap()
                }
                
                caddy_controller.upsert_site_configuration(&site.domain, &site.path.join(&active_version.id)).await?;
            }
        }
        
        if state.is_outdated() {
            download_tasks_sender.send(site.domain.clone()).unwrap()
        }
    }

    let join_handle = tokio::spawn(sites_worker(
        download_tasks_receiver,
        Arc::clone(&sites),
        caddy_controller,
    ));

    Ok(SitesWorker {
        sites,
        sites_directory_path,
        download_tasks_sender,
        join_handle,
    })
}
